/**/
#include "FloatingPoint.mqh"
#include "Types.mqh"
/****************************************************************
Arrays class
Frequent array operation methods.
Sometimes we need to add an element to array, or insert an element.
Standard Mql Array functions are efficient but do not provide such functionality.
For example, if you need to find an element, you can search only in sorted array with ArrayBsearch.
They also do not support complex data types.
This class fills the gap, and provides some frequent array operation methods:
   Add/insert elements to array by value/reference.
   Find/ReverseFind an element in unsorted array.
   Sum/Average of a numeric array.
   Min/Max returns the actual min/max value, not index of the value.
   Array copy/insert works with arrays containing complex types.
   StringToCodes creates an array of symbol codes from a string.
/****************************************************************/
class Arrays
  {
public:
   template<typename T>int    Add(T src,T &dst[]);
   template<typename T>int    Copy(T &dst[],T &srs[],bool clean=0,int dststart=0,int srcstart=0,int srccount=WHOLE_ARRAY);
   template<typename T>bool   Insert(T src,T &dst[],int pos);
   template<typename T>int    Find(T element,T &array[],int start=0);
   template<typename T>int    FindReverse(T element,T &array[],int start=0);
   template<typename T>double Average(T &src[]);
   template<typename T>double Sum(T &src[]);
   template<typename T>double Max(T& src[]);
   template<typename T>double Min(T& src[]);

   template<typename T>int    Addbyref(T &src,T &dst[]);
   template<typename T>bool   Insertbyref(T &src,T &dst[],int pos);
   template<typename T>bool   Insert(T &dst[],T &src[],int dststart);
   int                        StringToCodes(string src,short &dst[],int startpos=0,int count=-1);
protected:
   Types             types;
  };
/****************************************************************/
template<typename T>double Arrays::Min(T &src[])
  {
   if(!types.IsNumeric(typename(T)))
     {return -1.;}
   int id=ArrayMinimum(src);
   return (id==-1)?id:src[id];
  }
/****************************************************************/
template<typename T>double Arrays::Max(T &src[])
  {
   if(!types.IsNumeric(typename(T)))
     {return -1.;}
   int id=ArrayMaximum(src);
   return (id==-1)?id:src[id];
  }
/****************************************************************/
template<typename T>double Arrays::Sum(T &src[])
  {
   if(!types.IsNumeric(typename(T)))
     {return -1.;}
   double sum=0.;
   int size=ArraySize(src);
   for(int i=0; i<size; i++)
     {sum+=src[i];}
   return sum;
  }
/****************************************************************/
template<typename T>double Arrays::Average(T &src[])
  {
   if(!types.IsNumeric(typename(T)))
     {return -1.;}
   double sum=Sum(src);
   int size=ArraySize(src);
   return (sum/((size)?size:1));
  }
/****************************************************************/
template<typename T>bool Arrays::Insert(T &dst[],T &src[],int dststart)
  {
   /*CHECK*/
   int sizedst=ArraySize(dst);
   int sizesrc=ArraySize(src);
   if(dststart>sizedst||dststart<0)
     {PrintFormat("%s: dststart parameter out of range: %d",__FUNCTION__,dststart); return false;}
   /*INSERT ARRAY*/
   T tmp[];
   Copy(tmp,dst,0,0,0,dststart);
   Copy(tmp,src,0,dststart);
   Copy(tmp,dst,0,dststart+sizesrc,dststart);
   Copy(dst,tmp,1);
   return true;
  }
/****************************************************************/
template<typename T>bool Arrays::Insert(T element,T &array[],int pos)
  {
   /*CHECK*/
   if(pos>ArraySize(array)||pos<0)
     {PrintFormat("%s: pos parameter out of range: %d",__FUNCTION__,pos); return false;}
   /*INSERT ELEMENT*/
   T tmp[];
   Copy(tmp,array,0,0,0,pos);
   Add(element,tmp);
   Copy(tmp,array,0,pos+1,pos);
   Copy(array,tmp,1);
   return true;
  }
/****************************************************************/
template<typename T>bool Arrays::Insertbyref(T &element,T &array[],int pos)
  {
   if(pos>ArraySize(array)||pos<0)
     {PrintFormat("%s: pos parameter out of range: %d",__FUNCTION__,pos); return false;}
   /*INSERT ELEMENT BY REF*/
   T tmp[];
   Copy(tmp,array,0,0,0,pos);
   Addbyref(element,tmp);
   Copy(tmp,array,0,pos+1,pos);
   Copy(array,tmp,1);
   return true;
  }
/****************************************************************/
template<typename T>int Arrays::Add(T element,T &array[])
  {
   int size=ArraySize(array);
   if(!(ArrayResize(array,size+1,0x16)>-1))
     {Print(__FUNCTION__,": ArrayResize error: ",GetLastError());}
   else
     {array[size]=element;}
   return ArraySize(array);
  }
/****************************************************************/
template<typename T>int Arrays::Find(T element,T &array[],int start=0)
  {
   int ret=-1;
   /*CHECK*/
   int total=ArraySize(array);
   bool outofrange=start<0||start>total-1;
   if(outofrange)
     {PrintFormat("%s: start parameter out of range (%d)",__FUNCTION__,start); return -1;}
   /*SEARCH*/
   for(int i=start; i<total; i++)
     {
      T itm=array[i];
      bool elementisitem;
      if(types.IsFloatingPoint(typename(T)))
        {FloatingPoint<T>fp; elementisitem=fp.Compare(element,itm)==0;}
      else
        {elementisitem=element==itm;}
      if(elementisitem)
        {ret=i; break;}
     }
   return ret;
  }
/****************************************************************
FindReverse searches for the element from array end.
/****************************************************************/
template<typename T>int Arrays::FindReverse(T element,T &array[],int start=0)
  {
   T copy[];
   Copy(copy,array);
   if(!ArrayReverse(copy))
     {Print(__FUNCTION__,": ArrayReverse error: ",GetLastError()); return -1;}
   return Find(element,copy,start);
  }
/****************************************************************
Addbyref does the same thing as Add by can work with complex types,
 which are sent only by reference.
/****************************************************************/
template<typename T>int Arrays::Addbyref(T &element,T &array[])
  {
   int size=ArraySize(array);
   if(!(ArrayResize(array,size+1,0x10)>-1))
     {Print(__FUNCTION__,": ArrayResize error: ",GetLastError());}
   else
     {array[size]=element;}
   return ArraySize(array);
  }
/****************************************************************
Copy has the same signature/behavior as the standard Mql ArrayCopy.
Unlike ArrayCopy, it allows complex types in src/dst.
/****************************************************************/
template<typename T>int Arrays::Copy(T &dst[],T &srs[],bool clean=0,int dststart=0,int srcstart=0,int srccount=-1)
  {
   /*CLEAN*/
   if(clean)
     {ArrayFree(dst); dststart=0;}
   /*SIZES*/
   int sizesrc=ArraySize(srs);
   int sizedst=ArraySize(dst);
   /*CHECK*/
   bool dststartok=dststart>=0&&dststart<=sizedst;
   bool srcstartok=srcstart>=0&&srcstart<=sizesrc;
   bool wholearray=srccount==WHOLE_ARRAY;
   bool srccountok=(wholearray)?true:srccount+srcstart<=sizesrc;
   if(!dststartok||!srcstartok||!srccountok)
     {Print(__FUNCTION__,": parameters check error"); return -1;}
   bool srccountzero=srccount==0;
   if(srccountzero)
     {return 0;}
   /*RESIZE*/
   srccount=(wholearray)?sizesrc-srcstart:srccount;
   bool resize=dststart+srccount>sizedst;
   int sizedstnew=dststart+srccount;
   if(resize)
     {ArrayResize(dst,sizedstnew);}
   /*COPY*/
   int i,until=srccount;
   for(i=0; i<until; i++,dststart++,srcstart++)
     {dst[dststart]=srs[srcstart];}
   return i;
  }
/****************************************************************
StringToCodes is basically the standard StringToShortArray,
 except that the final 0 is removed from the destination array.
/****************************************************************/
int Arrays::StringToCodes(string src,short &dst[],int startpos=0,int count=-1)
  {
   if((StringToShortArray(src,dst,startpos,count)))
     {
      if(!(ArrayRemove(dst,ArraySize(dst)-1)))
        {Print(__FUNCTION__,": ArrayRemove error: ",GetLastError());}
     }
   return ArraySize(dst);
  }
/**/
